--------------------------------------------------------------------------------
-- Generic deriving with support for 1 parameter (GHC.Generics)
--------------------------------------------------------------------------------

{-# LANGUAGE ConstraintKinds        #-}
{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE DefaultSignatures      #-}
{-# LANGUAGE TypeOperators          #-}
{-# LANGUAGE UndecidableInstances   #-}
{-# LANGUAGE OverlappingInstances   #-}
{-# LANGUAGE IncoherentInstances    #-}
{-# LANGUAGE MultiParamTypeClasses  #-}

module Library.Deriving ( module Library.Deriving, PRU(..), SO(..), NoPar
                        , module Auxiliary.Metadata ) where


import Auxiliary.Metadata
import Auxiliary.Utils (NoPar)
import Control.Applicative
import Data.Foldable
import Prelude hiding (foldr)
import GHC.Exts ( Constraint )

--------------------------------------------------------------------------------
-- Un
--------------------------------------------------------------------------------

data Un s = 
  -- Void (used for datatypes without constructors)
    VV
    
  -- Unit
  | UU

  -- Meta-data (for constructors only, for now)
  | MM MetaCons (Un s)
  
  -- A parameter
  | PAR
  
  -- Constants (either other parameters or recursion into types of kind *)
  | KK PRU s

  -- Recursion into types of kind (* -> *)
  | REC SO (s -> s)
  
  -- Sum, product
  | Un s :+:  Un s
  | Un s :**: Un s

  -- Composition
  | (s -> s) :.: Un s

  -- Functions (not currently in the released version of the library)
  -- | FUN s (Un s)

-- Parameter, Recursive occurence, or Unknown/other
data PRU = P | R SO | U

-- Self or Other
data SO = S | O

--------------------------------------------------------------------------------
-- Interpretation (as a GADT)
--------------------------------------------------------------------------------

data In (u :: Un *) (p :: *) :: * where

  -- No interpretation for VV, as it shouldn't map to any value
  
  -- Unit
  U1     :: In UU p
  
  -- A constructor
  M1     :: In a p -> In (MM mc a) p

  -- The parameter
  Par1   :: { unPar1 :: p } -> In PAR p
  
  -- Constants
  K1     :: { unK1 :: x} -> In (KK pru x) p

  -- Recursion
  Rec1   :: { unRec1 :: f p } -> In (REC i f) p
  
  -- Sum
  L1     :: In f p -> In (f :+: g) p
  R1     :: In g p -> In (f :+: g) p

  -- Product
  (:*:)  :: In f p -> In g p -> In (f :**: g) p

  -- Composition
  Comp1  :: { unComp1 :: f (In g p) } -> In (f :.: g) p

  -- Functions
  -- Fun    :: { unFun :: a -> In f p } -> In (FUN a f) p

--------------------------------------------------------------------------------
-- Conversions to/from user datatypes
--------------------------------------------------------------------------------

class Generic (a :: *) where
  type Rep a :: Un *
  type Par a :: *
  type Par a = NoPar

  from :: a -> In (Rep a) (Par a)
  to   :: In (Rep a) (Par a) -> a


class Generic1 (f :: * -> *) where
  type Rep1 f :: Un *

  from1 :: f p -> In (Rep1 f) p
  to1   :: In (Rep1 f) p -> f p

--------------------------------------------------------------------------------
-- Example function: equality
--------------------------------------------------------------------------------

class GEq' (f :: Un *) where
  geq' :: In f a -> In f a -> Bool

instance GEq' UU where
  geq' _ _ = True

instance (GEq c) => GEq' (KK i c) where
  geq' (K1 a) (K1 b) = geq a b

instance (GEq' a) => GEq' (MM c a) where
  geq' (M1 a) (M1 b) = geq' a b

-- No instances for P or Rec because geq is only applicable to types of kind *

instance (GEq' a, GEq' b) => GEq' (a :+: b) where
  geq' (L1 a) (L1 b) = geq' a b
  geq' (R1 a) (R1 b) = geq' a b
  geq' _      _      = False

instance (GEq' a, GEq' b) => GEq' (a :**: b) where
  geq' (a1 :*: b1) (a2 :*: b2) = geq' a1 a2 && geq' b1 b2

instance (Applicative f, Foldable f, GEq' a) => GEq' (f :.: a) where
  -- I'm sure there's a simpler way to do this, but hey...
  geq' (Comp1 x) (Comp1 y) = length (toList x) == length (toList y)
                          && foldr (&&) True (geq' <$> x <*> y)

-- instance GEq' (FUN a f) where
  -- geq' _ _ = False


-- User-facing class
class GEq (a :: *) where 
  geq :: a -> a -> Bool

  default geq :: (Generic a, GEq' (Rep a)) => a -> a -> Bool
  geq x y = geq' (from x) (from y)
  -- geq = generic geq'

-- Base types instances
instance GEq Char   where geq = (==)
instance GEq Int    where geq = (==)
instance GEq Float  where geq = (==)

--------------------------------------------------------------------------------
-- Example function: fmap
--------------------------------------------------------------------------------

class GFunctor' (f :: Un *) where
  gmap' :: (a -> b) -> In f a -> In f b

instance GFunctor' UU where
  gmap' _ U1 = U1

instance GFunctor' PAR where
  gmap' f (Par1 a) = Par1 (f a)

instance GFunctor' (KK i c) where
  gmap' _ (K1 a) = K1 a

instance (GFunctor f) => GFunctor' (REC i f) where
  gmap' f (Rec1 a) = Rec1 (gmap f a)

instance (GFunctor' f) => GFunctor' (MM i f) where
  gmap' f (M1 a) = M1 (gmap' f a)

instance (GFunctor' f, GFunctor' g) => GFunctor' (f :+: g) where
  gmap' f (L1 a) = L1 (gmap' f a)
  gmap' f (R1 a) = R1 (gmap' f a)

instance (GFunctor' f, GFunctor' g) => GFunctor' (f :**: g) where
  gmap' f (a :*: b) = gmap' f a :*: gmap' f b

instance (GFunctor f, GFunctor' g) => GFunctor' (f :.: g) where
  gmap' f (Comp1 x) = Comp1 (gmap (gmap' f) x)

-- instance (GFunctor' f) => GFunctor' (FUN a f) where
  -- gmap' f (Fun g) = Fun (gmap' f . g)


-- User-facing class
class GFunctor (f :: * -> *) where
  gmap :: (a -> b) -> f a -> f b

  default gmap :: (Generic1 f, GFunctor' (Rep1 f))
               => (a -> b) -> f a -> f b
  gmap f = to1 . gmap' f . from1
  -- gmap f = generic (gmap' f)
